searchNodes=[{"doc":"Authorization superpowers for applications using Ecto . Priorities: Single source of truth - The same rules that authorize loaded data should be able to load authorized data. Authentication-agnostic - Janus should not care about how users are modeled or authenticated. Minimal library footprint - Expose a small but flexible API that can be used to create an optimal authorization interface for each application. Escape hatches where necessary - Complex authorization rules and use-cases should be representable when Janus neglects to provide a short cut. Janus is split into two primary components: Janus.Policy - functions and behaviour for defining policy modules , which describe the allowed actors, actions, and resources in your application. This is where you look if you're writing a policy module. Janus.Authorization - functions and behaviour used by the rest of your application to authorize and load resources. This is where you look if you're using a policy module. Janus defines a Mix task to generate the basic policy module that will get you started: $ mix janus.gen.policy Installation Janus can be installed by adding ex_janus to your deps in mix.exs : defp deps do [ { :ex_janus , &quot;~&gt; 0.3.0&quot; } ] end Policies Policy modules are created by invoking use Janus , which implements both the Janus.Policy and Janus.Authorization behaviours: defmodule Policy do use Janus @impl true def build_policy ( policy , _actor ) do policy end end When you invoke use Janus , default implementations are injected for required callbacks, except for Janus.Policy.build_policy/2 . This callback is your foundation, as it returns the authorization policy for an individual user of your application. The policy above is not very useful (it doesn't allow anyone to do anything) but that can be changed by using the Janus.Policy API to define actions, resources, and conditions that make up your authorization rules. def build_policy ( policy , % User { role : :moderator } = mod ) do policy |&gt; allow ( Post , :read ) |&gt; allow ( Post , [ :edit , :archive , :unarchive ] , where : [ user : [ role : :member ] ] ) |&gt; allow ( Post , [ :edit , :archive , :unarchive ] , where : [ user_id : mod . id ] ) |&gt; deny ( Post , :unarchive , where : [ archived_by : [ role : :admin ] ] ) end See the Janus.Policy documentation for more on defining policies. Authorization With our policy module defined, it can now be used to load and authorize resources. iex&gt; Policy . authorize ( some_post , :archive , moderator ) { :ok , some_post } iex&gt; Policy . authorize ( post_archived_by_admin , :unarchive , moderator ) { :error , :not_authorized } iex&gt; Policy . scope ( Post , :read , moderator ) % Ecto.Query { } iex&gt; Policy . scope ( Post , :read , moderator ) |&gt; Repo . all ( ) [ ... posts the moderator can read ] iex&gt; Policy . any_authorized? ( Post , :edit , moderator ) true # there are rules allowing moderators to edit posts iex&gt; Policy . any_authorized? ( Post , :delete , moderator ) false # there are no rules that allow moderators to delete posts These functions make up the Janus.Authorization behaviour, and their definitions were injected by default when we invoked use Janus . This is the &quot;public API&quot; that the rest of your application will use to authorize resources. See the Janus.Authorization documentation for more. Integration with Ecto.Query The primary assumption that Janus makes is that your resources are backed by an Ecto.Schema . Using Ecto's schema reflection capabilities, Janus is able to use the same policy to authorize a single resource and to construct a composable Ecto query that is aware of field types and associations. # This query would result in the 5 latest posts that the current # user is authorized to see, preloaded with the user # who made # the post (but only if the current user is allowed to see that # user). Post |&gt; Policy . scope ( :read , current_user , preload_authorized : :user ) |&gt; order_by ( desc : :inserted_at ) |&gt; limit ( 5 ) This integration with Ecto queries is main reason Janus exists. Configuration Some defaults can be configured by passing them as options when invoking use Janus . Those are: :repo - Ecto.Repo used to load associations when required by your authorization rules :load_associations - Load associations when required by your authorization rules (requires :repo config option to be set or to be passed explicitly at the call site), defaults to false For example: defmodule MyApp.Policy do use Janus , repo : MyApp.Repo , load_associations : true # ... end These defaults will be referenced in the Janus.Authorization documentation where they are used. Why (not) Janus? Janus was created to scratch an itch: the same rules that authorize loaded data should be able to load authorized data. In concrete terms, a rule that defines whether a user can edit a resource should also be able to load all the resources that user can edit. Loading data this way should be: efficient - loading everything and then filtering it in-memory doesn't cut it; composable - it should be possible to add additional conditions when loading data; ergonomic - authorization should slot-in naturally without major rewrites. Thankfully, integration with Ecto.Query solves for all of the above. One only needs authorization rules that can be translated into a query. And thus, Janus was born. Janus may be a good fit if... you're authorizing data backed by Ecto.Schema . Janus relies on the reflection capabilities of schemas to produce correct queries, cast values, navigate associations, etc. you share interfaces between users with different permissions. Janus allows you to scope queries in a uniform way using the current user (or lack of one), making shared interfaces a natural default. you prefer to have the final say. Janus takes an approach similar to Phoenix, generating code that supports certain conventions while allowing you to override or redefine behavior to fit your preferences. you prefer a functional API for defining rules. Authorization policies are data; adding an authorization rule just transforms that data. Policies can be built using the full extent and natural composability of the Elixir language. Janus may not be a good fit if... you're only authorizing actions that don't have an obvious association to data backed by Ecto.Schema . For instance, a :send_welcome_email action without some kind of Email schema. Janus does, however, give you a natural place to define that sort of API yourself (your policy module). you want an easy-to-read DSL for authorization rules. Janus policies are &quot;just code&quot;, so readability will depend on your own style and structure. If you value readability/scannability very highly, definitely check out LetMe , which provides a great DSL and makes some different trade-offs than Janus does. you want runtime introspection for your authorization rules, like a list of all actions a user can perform. Janus does not currently provide structured access to this information, but you might again turn to LetMe , which provides introspection capabilities.","ref":"Janus.html","title":"Janus","type":"module"},{"doc":"Sets up a module to implement the Janus.Policy and Janus.Authorization behaviours. Using use Janus does the following: adds the Janus.Policy behaviour, imports functions used to define the required callback Janus.Policy.build_policy/2 , and defines a build_policy/1 helper adds the Janus.Authorization behaviour and injects default (overridable) implementations for all callbacks Options :load_associations - Load associations when required by your authorization rules (requires :repo config option to be set or to be passed explicitly at the call site), defaults to false :repo - Ecto.Repo used to load associations when required by your authorization rules See &quot;Configuration&quot; section for details. Example defmodule MyApp.Policy do use Janus , repo : MyApp.Repo @impl true def build_policy ( policy , _actor ) do policy # |&gt; allow(...) end end","ref":"Janus.html#__using__/1","title":"Janus.__using__/1","type":"macro"},{"doc":"","ref":"Janus.html#t:action/0","title":"Janus.action/0","type":"type"},{"doc":"","ref":"Janus.html#t:actor/0","title":"Janus.actor/0","type":"type"},{"doc":"","ref":"Janus.html#t:schema_module/0","title":"Janus.schema_module/0","type":"type"},{"doc":"Authorize and load resources using policies. Policy modules expose a minimal API that can be used to authorize and load resources throughout the rest of your application. authorize/4 - authorize an individual, already-loaded resource scope/4 - construct an Ecto query for a schema that will filter results to only those that are authorized any_authorized?/3 - checks whether the given actor/policy has any access to the given schema for the given action These functions will usually be called from your policy module directly, since wrappers that accept either a policy or an actor are injected when you invoke use Janus . Documentation examples will show usage from your policy module. See individual function documentation for details.","ref":"Janus.Authorization.html","title":"Janus.Authorization","type":"behaviour"},{"doc":"Checks whether any permissions are defined for the given schema, action, and actor. This function is most useful in conjunction with scope/4 , which builds an Ecto query that filters to only those resources the actor is authorized for. If you run the resulting query and receive [] , it is not possible to determine whether the result is empty because the actor wasn't authorized for any resources or because of other restrictions on the query. For example, you might use the following pattern to load all the resources a user is allowed to read that were inserted in the last day: query = from ( r in MyResource , where : r . inserted_at &gt; from_now ( - 1 , &quot;day&quot; ) ) if any_authorized? ( query , :read , user ) do { :ok , scope ( query , :read , user ) |&gt; Repo . all ( ) } else { :error , :not_authorized } end This would result in {:ok, results} if the user is authorized to read any resources, even if the result set is empty, and would result in {:error, :not_authorized} if the user isn't authorized to read the resources at all. Examples iex&gt; MyPolicy . any_authorized? ( MyResource , :read , actor ) true iex&gt; MyPolicy . any_authorized? ( MyResource , :delete , actor ) false","ref":"Janus.Authorization.html#any_authorized?/3","title":"Janus.Authorization.any_authorized?/3","type":"function"},{"doc":"","ref":"Janus.Authorization.html#c:any_authorized?/3","title":"Janus.Authorization.any_authorized?/3","type":"callback"},{"doc":"Authorizes a loaded resource. Expects to receive a struct, an action, and an actor or policy. Returns {:ok, resource} if authorized, otherwise {:error, :not_authorized} . Options :load_associations - Whether to load associations required by policy authorization rules, defaults to false unless configured on your policy module :repo - Ecto repository to use when loading required associations if :load_associations is set to true , defaults to nil unless configured on your policy module Examples iex&gt; MyPolicy . authorize ( % MyResource { } , :read , actor ) # accepts an actor { :ok , % MyResource { } } iex&gt; MyPolicy . authorize ( % MyResource { } , :read , policy ) # or a policy { :ok , % MyResource { } } iex&gt; MyPolicy . authorize ( % MyResource { } , :delete , actor ) { :error , :not_authorized }","ref":"Janus.Authorization.html#authorize/4","title":"Janus.Authorization.authorize/4","type":"function"},{"doc":"","ref":"Janus.Authorization.html#c:authorize/4","title":"Janus.Authorization.authorize/4","type":"callback"},{"doc":"Create an %Ecto.Query{} that results in only authorized records. Like the Ecto.Query API, this function can accept a schema as the first argument or a query, in which case it will compose with that query. If a query is passed, the appropriate schema will be derived from that query's source. scope ( MyResource , :read , user ) query = from ( r in MyResource , where : r . inserted_at &gt; from_ago ( 1 , &quot;day&quot; ) ) scope ( query , :read , user ) If the query specifies the source as a string, we cannot derive the schema. For example, this will not work: # Raises an ArgumentError query = from ( r in &quot;my_resources&quot; , where : r . inserted_at &gt; from_ago ( 1 , &quot;day&quot; ) ) scope ( query , :read , user ) Options :preload_authorized - Similar to Ecto.Query.preload/3 , but only preloads those associated records that are authorized. Note that this requires Ecto v3.9.4 or later and a database that supports lateral joins. See &quot;Preloading authorized associations&quot; for more information. Preloading authorized associations The :preload_authorized option can be used to preload associated records, but only those that are authorized for the given actor. An additional query can be specified for each preloaded association that will be run as if scoped to its parent row. This can simplify certain queries dramatically. For instance, imagine a user search interface that lists users along with their most recent comment. Naughty comments can be hidden by moderators, but those hidden comments should still be visible if a moderator is searching. Here's how that might be accomplished: iex&gt; last_comment = from ( Comment , order_by : [ desc : :inserted_at ] , limit : 1 ) iex&gt; User ...&gt; |&gt; search ( search_params ) ...&gt; |&gt; MyPolicy . scope ( :read , current_user , ...&gt; preload_authorized : [ comments : last_comment ] ...&gt; ) ...&gt; |&gt; Repo . all ( ) [ % User { comments : [ % Comment { } ] } , % User { comments : [ % Comment { } ] } , ... ] Some things to note about this example: The last_comment query runs as if scoped to each user's comments. This means that the :limit applies to each user's comments, not the entire set of comments. The comment will be the last inserted comment that is authorized to be read by the current_user . Moderators may be able to see hidden comments, while normal users may not. It is also possible to nest authorized preloads. For instance, you could preload comments and their associated post. MyPolicy . scope ( User , :read , current_user , preload_authorized : [ comments : :post ] ) This would load all comments. You could incorporate the last_comment query above by specifying it as the first element of a tuple, followed by the list of inner preloads: MyPolicy . scope ( User , :read , current_user , preload_authorized : [ comments : { last_comment , [ :post ] } ] ) This would load only the latest comment as well as its associated post (assuming it too is authorized to be read by current_user ). Examples iex&gt; MyPolicy . scope ( MyResource , :read , actor ) % Ecto.Query { } iex&gt; MyPolicy . scope ( MyResource , :read , actor ) |&gt; Repo . all ( ) [ % MyResource { } , ... ] iex&gt; MyResource ...&gt; |&gt; MyPolicy . scope ( :read , actor ) ...&gt; |&gt; order_by ( inserted_at : :desc ) ...&gt; |&gt; limit ( 1 ) ...&gt; |&gt; Repo . one ( ) % MyResource { } iex&gt; MyResource ...&gt; |&gt; MyPolicy . scope ( :read , actor , ...&gt; preload_authorized : :other ...&gt; ) ...&gt; |&gt; Repo . all ( ) [ % MyResource { other : % OtherResource { } } , ... ]","ref":"Janus.Authorization.html#scope/4","title":"Janus.Authorization.scope/4","type":"function"},{"doc":"","ref":"Janus.Authorization.html#c:scope/4","title":"Janus.Authorization.scope/4","type":"callback"},{"doc":"","ref":"Janus.Authorization.html#t:filterable/0","title":"Janus.Authorization.filterable/0","type":"type"},{"doc":"Define composable authorization policies for actors in your system. A policy is a data structure created for an actor in your system that defines the schemas that actor can access, the actions they can take, and any restrictions to the set of resources that can be accessed. These policies are generally created implicitly for actors passed to functions defined by Janus.Authorization , but they can also be created with build_policy/2 . Creating a policy modules While you can create a policy module with use Janus.Policy , you will usually invoke use Janus and implement build_policy/2 : defmodule MyApp.Policy do use Janus @impl true def build_policy ( policy , _actor ) do policy end end An implementation for build_policy/1 is injected into the policy module. Policy modules can now be used to generate policy structs explicitly (though they will usually be created implicitly when calling functions defined by Janus.Authorization ). iex&gt; policy = MyApp.Policy . build_policy ( :my_user ) % Janus.Policy { actor : :my_user , rules : %{ ... } } iex&gt; MyApp.SecondaryPolicy . build_policy ( policy ) % Janus.Policy { actor : :my_user , rules : %{ ... } } Permissions with allow and deny Permissions are primarily defined using allow/4 and deny/4 , which allows or denies an action on a resource if a set of conditions match. Both functions take the same arguments and options. When permissions are being checked, multiple allow rules combine using logical-or, with deny rules overriding allow . For example, the following policy would allow a moderator to edit their own comments and any comments flagged for review, but not those made by an admin. @impl true def build_policy ( policy , % User { role : :moderator } = user ) do policy |&gt; allow ( Comment , :update , where : [ user : [ id : user . id ] ] ) |&gt; allow ( Comment , :update , where : [ flagged_for_review : true ] ) |&gt; deny ( Comment , :update , where : [ user : [ role : :admin ] ] ) end While set of keyword options passed to allow and deny are reminiscent of keyword-based Ecto queries, but since they are functions and not macros, there is no need to use the ^value syntax used in Ecto. For example, the following would result in an error: allow ( policy , Comment , :update , where : [ user : [ id : ^ user . id ] ] ) :where and :where_not conditions These conditions match if the associated fields are equal to each other. For instance, the moderation example above could also be written as: @impl true def build_policy ( policy , % User { role : :moderator } = user ) do policy |&gt; allow ( Comment , :update , where : [ user_id : user . id ] ) |&gt; allow ( Comment , :update , where : [ flagged_for_review : true ] , where_not : [ user : [ role : :admin ] ] ) end Multiple conditions within the same allow / deny are combined with a logical-and, so this might be translated to English as &quot;allow moderators to edit comments they made or to edit comments flagged for review that were not made by an admin&quot;. :or_where conditions You can also use :or_where to combine with all previous conditions. For instance, the two examples above could also be written as: @impl true def build_policy ( policy , % User { role : :moderator } = user ) do policy |&gt; allow ( Comment , :update , where : [ flagged_for_review : true ] , where_not : [ user : [ role : :admin ] ] , or_where : [ user_id : user . id ] ) end An :or_where condition applies to all clauses before it. Using some pseudocode for demonstration, the above would read: # (flagged_for_review AND NOT user.role == :admin) OR user_id == user.id These clauses could be reordered to have a different meaning: policy |&gt; allow ( Comment , :update , where : [ flagged_for_review : true ] , or_where : [ user_id : user . id ] , where_not : [ user : [ role : :admin ] ] ) # (flagged_for_review OR user_id == user.id) AND NOT user.role == :admin Attribute checks with functions When equality is not a sufficient check for an attribute, a function can be supplied. For instance, a published_at field might be used to schedule posts. Users may only have permission to read posts where published_at is in the past, but we can only check for equality using the basic keyword syntax presented above. In these cases, you can defer this check using an arity-3 function: @impl true def build_policy ( policy , _actor ) do policy |&gt; allow ( Comment , :read , where : [ published_at : &amp; in_the_past? / 3 ] ) end def in_the_past? ( :boolean , record , :published_at ) do if value = Map . get ( record , :published_at ) do DateTime . compare ( DateTime . utc_now ( ) , value ) == :gt end end def in_the_past? ( :dynamic , binding , :published_at ) do now = DateTime . utc_now ( ) Ecto.Query . dynamic ( ^ now &gt; as ( ^ binding ) . published_at ) end As seen in the example above, functions must define at least two clauses based on their first argument, :boolean or :dynamic , so that they can handle both operations on a single record and operations that should compose with an Ecto query. Working with rulesets Policies can also be defined by attaching rulesets created using allow/3 and deny/3 . Instead of taking a policy as a first argument, these functions take a schema (or a ruleset). Rulesets are specific to an individual schema and can be attached to a policy using attach/2 . For example: @impl true def build_policy ( policy , actor ) do policy |&gt; attach ( rules_for ( Thread , actor ) ) |&gt; attach ( rules_for ( Post , actor ) ) end defp rules_for ( Thread , % User { id : user_id } ) do Thread |&gt; allow ( :read , where : [ archived : false ] ) |&gt; allow ( [ :create , :update ] , where : [ creator_id : user_id ] ) end defp rules_for ( Thread , nil ) do Thread |&gt; allow ( :read , where : [ archived : false , visibility : :public ] ) end defp rules_for ( Post , _actor ) do Post |&gt; allow ( :read , where : [ thread : allows ( :read ) ] ) end Depending on your specific needs, rulesets may allow you to organize policies in a way that is easier to maintain. In the above example, delegating to a private rules_for/2 function that returns a ruleset allows us to pattern-match on a nil user where it matters and share a ruleset where it doesn't. This pattern has tradeoffs, however. You would need to ensure that the pattern-matching for each schema is exhaustive, for instance, otherwise a FunctionClauseError might be raised. Hooks Functions can be registered as hooks that run prior to authorization calls. See attach_hook/4 for more information.","ref":"Janus.Policy.html","title":"Janus.Policy","type":"behaviour"},{"doc":"Creates or updates a ruleset for a schema to allow an action if matched by conditions. Must be attached to a policy using attach/2 . See &quot;Permissions with allow and deny &quot; for a description of conditions. Examples thread_rules = Thread |&gt; allow ( :read ) |&gt; allow ( :create , where : [ creator_id : user . id ] ) attach ( policy , thread_rules )","ref":"Janus.Policy.html#allow/3","title":"Janus.Policy.allow/3","type":"function"},{"doc":"Allows an action on the schema if matched by conditions. See &quot;Permissions with allow and deny &quot; for a description of conditions. Examples policy |&gt; allow ( FirstResource , :read ) |&gt; allow ( SecondResource , :create , where : [ creator : [ id : user . id ] ] )","ref":"Janus.Policy.html#allow/4","title":"Janus.Policy.allow/4","type":"function"},{"doc":"Specifies that a condition should match if another action is allowed. If used as the value for an association, the condition will match if the action is allowed for the association. Examples Allow users to edit any posts they can delete. policy |&gt; allow ( Post , :update , where : allows ( :delete ) ) |&gt; allow ( Post , :delete , where : [ user_id : user . id ] ) Don't allow users to edit posts they can't read. policy |&gt; allow ( Post , :read , where : [ archived : false ] ) |&gt; allow ( Post , :update , where : [ user_id : user . id ] ) |&gt; deny ( Post , :update , where_not : allows ( :read ) ) Example with associations Let's say we have some posts with comments. Posts are visible unless they are archived, and all comments of visible posts are also visible. To start, we can duplicate the condition: policy |&gt; allow ( Post , :read , where : [ archived : false ] ) |&gt; allow ( Comment , :read , where : [ post : [ archived : false ] ] ) If we add additional clauses to the condition for posts, however, we will have to duplicate them for comments. We can use allows instead: policy |&gt; allow ( Post , :read , where : [ archived : false ] ) |&gt; allow ( Comment , :read , where : [ post : allows ( :read ) ] ) Now let's say we add a feature that allows for draft posts, which should not be visible unless a published_at is set. We can modify only the condition for Post and that change will propogate to comments. policy |&gt; allow ( Post , :read , where : [ archived : false ] , where_not : [ published_at : nil ] ) |&gt; allow ( Comment , :read , where : [ post : allows ( :read ) ] )","ref":"Janus.Policy.html#allows/1","title":"Janus.Policy.allows/1","type":"function"},{"doc":"Attach a ruleset created using allow/3 and deny/3 to a policy. Examples thread_rules = Thread |&gt; allow ( :read ) |&gt; deny ( :read , where : [ scope : :private ] ) attach ( policy , thread_rules )","ref":"Janus.Policy.html#attach/2","title":"Janus.Policy.attach/2","type":"function"},{"doc":"Attach a hook to the policy. Expects the following arguments: policy - the %Janus.Policy{} struct to attach to name - an atom identifying the hook schema (default :all ) - an Ecto schema module identifying the resource or query source that the hook should be applied to fun - the hook function, see &quot;Hooks&quot; below If the given name is already present, an error will be raised. If you wish to replace a hook, you can use detach_hook/3 before re-attaching the hook. If you only wish to add a hook if it is hasn't already been added, use attach_new_hook/4 instead. Hooks will be run in the order that they are attached. Hooks Hooks are anonymous or captured functions that accept three arguments: operation - one of :authorize or :scope object - either a struct (for :authorize ) or a queryable (for :scope ) that is being authorized action - the action being authorized Hooks must return one of the following: {:cont, object} - additional hooks and authorization continue :halt - halt authorization, running no additional hooks and returning {:error, :not_authorized} for an :authorize operation and an empty query for a :scope operation Examples When writing hooks, you must ensure that all possible arguments are handled. This can be done using a &quot;catch-all&quot; clause. For example: policy |&gt; attach_hook ( :preload_user , fn :authorize , % Post { } = resource , _action -&gt; { :cont , Repo . preload ( resource , :user ) } _operation , object , _action -&gt; { :cont , object } end ) policy |&gt; attach_hook ( :preload_user , Post , fn :authorize , resource , _action -&gt; { :cont , Repo . preload ( resource , :user ) } _operation , object , _action -&gt; { :cont , object } end ) Hooks can also be captured functions. For example: policy |&gt; attach_hook ( :preload_user , Post , &amp; preload_user / 3 ) # elsewhere in your module defp preload_user ( :authorize , resource , _action ) do { :cont , Repo . preload ( resource , :user ) } end defp preload_user ( :scope , query , _action ) do { :cont , from ( query , preload : :user ) } end If :halt is returned from a hook, no further hooks will be run and nothing will be authorized. This could be used to perform a check on banned users, for example: @impl true def build_policy ( policy , user ) do policy |&gt; attach_hook ( :ensure_unbanned , fn _op , object , _action -&gt; if Accounts . banned? ( user . id ) do :halt else { :cont , object } end end ) end This may be required if policies are being cached, since the hook runs every time the authorization call happens, instead of only once when the policy is built.","ref":"Janus.Policy.html#attach_hook/4","title":"Janus.Policy.attach_hook/4","type":"function"},{"doc":"Attach a new hook to the policy. Like attach_hook/4 , except it only attaches the hook if the name isn't present for the given schema.","ref":"Janus.Policy.html#attach_new_hook/4","title":"Janus.Policy.attach_new_hook/4","type":"function"},{"doc":"Builds an authorization policy, delegating to build_policy/2 . If given a policy, calls build_policy/2 with the policy and the actor associated with the policy. If given an actor, creates an empty policy associated with that actor and passes it to build_policy/2 . An implementation for this callback is injected into modules invoking either use Janus or use Janus.Policy .","ref":"Janus.Policy.html#c:build_policy/1","title":"Janus.Policy.build_policy/1","type":"callback"},{"doc":"Builds an authorization policy containing rules for the given actor. See Janus.Policy for API documentation on building policies.","ref":"Janus.Policy.html#c:build_policy/2","title":"Janus.Policy.build_policy/2","type":"callback"},{"doc":"Creates or updates a ruleset for a schema to deny an action if matched by conditions. Must be attached to a policy using attach/2 . See &quot;Permissions with allow and deny &quot; for a description of conditions. Examples thread_rules = Thread |&gt; allow ( :read ) |&gt; deny ( :read , where : [ scope : :private ] ) attach ( policy , thread_rules )","ref":"Janus.Policy.html#deny/3","title":"Janus.Policy.deny/3","type":"function"},{"doc":"Denies an action on the schema if matched by conditions. See &quot;Permissions with allow and deny &quot; for a description of conditions. Examples policy |&gt; allow ( FirstResource , :read ) |&gt; deny ( FirstResource , :read , where : [ scope : :private ] )","ref":"Janus.Policy.html#deny/4","title":"Janus.Policy.deny/4","type":"function"},{"doc":"Detach a hook from the policy.","ref":"Janus.Policy.html#detach_hook/3","title":"Janus.Policy.detach_hook/3","type":"function"},{"doc":"","ref":"Janus.Policy.html#t:hook/0","title":"Janus.Policy.hook/0","type":"type"},{"doc":"","ref":"Janus.Policy.html#t:ruleset/0","title":"Janus.Policy.ruleset/0","type":"type"},{"doc":"","ref":"Janus.Policy.html#t:t/0","title":"Janus.Policy.t/0","type":"type"},{"doc":"Creates a basic policy module. $ mix janus.gen.policy [--module MODULE] [--path PATH] Creates the following file: lib/app_name/policy.ex - A Janus policy module containing additional authorization helpers (see Generated Policy Modules for more info) Options --module - The name of the generated module, defaults to AppName.Policy --app - The name of the application namespace, defaults to your application name camelized, e.g. AppName --path - The path (including filename) for the generated module, defaults to lib/app_name/policy.ex","ref":"Mix.Tasks.Janus.Gen.Policy.html","title":"Mix.Tasks.Janus.Gen.Policy","type":"task"},{"doc":"Defines a rule for an individual schema and action.","ref":"Janus.Policy.Rule.html","title":"Janus.Policy.Rule","type":"module"},{"doc":"","ref":"Janus.Policy.Rule.html#t:t/0","title":"Janus.Policy.Rule.t/0","type":"type"},{"doc":"This project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Enhancements [Janus.Policy] Add rulesets: Janus.Policy.attach/2 , Janus.Policy.allow/3 , Janus.Policy.deny/3 . [mix janus.gen.policy] Simplify generated module and function docs. Fixes [Janus.Policy] Fix validation that was incorrectly raising when rules were defined for a schema module whose code hadn't yet been loaded.","ref":"changelog.html#v0-3-1-2023-01-17","title":"Changelog - v0.3.1 (2023-01-17)","type":"extras"},{"doc":"Enhancements [ Breaking / Janus.Policy] Remove :module field from %Janus.Policy{} struct. [ Breaking / Janus.Policy] Change schema/action argument order in allow/4 and deny/4 to be more consistent with the rest of Janus. See example below. [Janus.Policy] Add Janus.Policy.build_policy/1 callback primarily to document its usage; an implementation was already injected into policy modules. [Janus.Policy] Add :actor field to %Janus.Policy{} struct. Example allow/4 and deny/4 # Old argument order policy |&gt; allow ( :read , Thread , where : [ ... ] ) |&gt; allow ( :create , Thread , where : [ ... ] ) |&gt; deny ( :create , Thread , where : [ ... ] ) # New argument order policy |&gt; allow ( Thread , :read , where : [ ... ] ) |&gt; allow ( Thread , :create , where : [ ... ] ) |&gt; deny ( Thread , :create , where : [ ... ] )","ref":"changelog.html#v0-3-0-2023-01-16","title":"Changelog - v0.3.0 (2023-01-16)","type":"extras"},{"doc":"Enhancements [Janus.Policy] Raise if something other than a schema is passed to allow/4 or deny/4 .","ref":"changelog.html#v0-2-1-2023-01-14","title":"Changelog - v0.2.1 (2023-01-14)","type":"extras"},{"doc":"Enhancements [ Breaking / Janus.Policy] policy_for renamed to build_policy . [ Breaking / Janus.Policy] before_policy_for removed. [ Breaking / Janus.Authorization] filter_authorized renamed to scope . [ Breaking / Janus.Authorization] authorize now returns {:error, :not_authorized} instead of :error . [Janus] use Janus accepts optional defaults for Janus.Authorization.authorize/4 options. [Janus.Policy] Pre-authorization hooks with Janus.Policy.attach_hook/4 and friends. [Janus.Authorization] authorize/4 adds :repo and :load_assocations options. [mix janus.gen.policy] Overhaul generated policy helpers to more easily replace Ecto.Repo callbacks.","ref":"changelog.html#v0-2-0-2023-01-05","title":"Changelog - v0.2.0 (2023-01-05)","type":"extras"},{"doc":"First release.","ref":"changelog.html#v0-1-0-2022-12-25","title":"Changelog - v0.1.0 (2022-12-25)","type":"extras"},{"doc":"","ref":"basics.html","title":"The Basics","type":"extras"},{"doc":"Generator $ mix janus.gen.policy * creating lib/my_app/policy.ex Generate a starting policy module for your application. For more information, see the Generated Policy Modules guide. Policy module Your policy module is the interface used by the rest of your application. This is usually the only place you should be referring to Janus directly. lib/my_app/policy.ex defmodule MyApp.Policy do use Janus @impl true def build_policy ( policy , actor ) do policy |&gt; # authorization rules end end For more information on defining policies, see the Policy Definition Cheatsheet","ref":"basics.html#setup","title":"The Basics - Setup","type":"extras"},{"doc":"Schemas Schemas are modules that use Ecto.Schema . Used when defining policies policy |&gt; allow ( Post , :edit , ... ) # ^^^^ Used when scoping a query MyApp.Policy . scope ( Post , :edit , current_user ) # ^^^^ Used when checking for permissions MyApp.Policy . any_authorized? ( Post , :edit , current_user ) # ^^^^ Resources Resources are loaded structs defined by one of your schemas. Used when authorizing an action MyApp.Policy . authorize ( % Post { } , :edit , current_user ) # ^^^^^^^ Actors Actors are the users of your application. They can be a %User{} struct, but they don't have to be. Actors are converted to a policy using build_policy/2 , so an actor can be anything that you want to use to differentiate between types of user. They can even be a simple atom like :normal_user or :admin_user . In build_policy/2 def build_policy ( policy , % User { } ) do # ^^^^^^^ end Used when calling any authorization function MyApp.Policy . authorize ( % Post { } , :edit , current_user ) # ^^^^^^^^^^^^ MyApp.Policy . scope ( Post , :edit , current_user ) # ^^^^^^^^^^^^ Actions Actions are what actors do to resources in your application. Janus doesn't care how you represent actions, but atoms usually do the trick. Used when defining policies policy |&gt; allow ( Post , :edit , ... ) # ^^^^^ Used when calling any authorization function MyApp.Policy . authorize ( % Post { } , :edit , current_user ) # ^^^^^ Can be any term except a list policy |&gt; allow ( Post , :edit , ... ) |&gt; allow ( Post , &quot;edit&quot; , ... ) |&gt; allow ( Post , % Action { type : :edit } , ... ) # lists are special-cased to allow multiple # actions to share conditions |&gt; allow ( Post , [ :read , :edit ] , ... )","ref":"basics.html#core-concepts","title":"The Basics - Core concepts","type":"extras"},{"doc":"Overview Authorization rules are attached to policies in the build_policy/2 callback. Janus.Policy.allow/4 Janus.Policy.deny/4 Janus.Policy.allows/1 Grant permission for all resources of schema policy |&gt; allow ( Post , :read ) |&gt; allow ( Post , :edit ) |&gt; allow ( Post , :archive ) |&gt; allow ( Comment , :read ) |&gt; allow ( Comment , :edit ) Using lists of actions policy |&gt; allow ( Post , [ :read , :edit , :archive ] , Post ) |&gt; allow ( Comment , [ :read , :edit ] , Comment ) Grant permission based on attributes policy |&gt; allow ( Post , :read , where : [ archived : false ] ) # or define using :where_not |&gt; allow ( Post , :read , where_not : [ archived : true ] ) # or override a blanket permission using deny |&gt; allow ( Post , :read ) |&gt; deny ( Post , :read , where : [ archived : true ] ) Use deny to override a previous allow policy |&gt; allow ( Post , :read ) |&gt; deny ( Post , :read , where : [ archived : true ] ) Grant permission if the user is associated with the resource def build_policy ( policy , % User { role : :member } = user ) do policy |&gt; allow ( Comment , :edit , where : [ user_id : user . id ] ) end Grant permission based on association attributes policy |&gt; allow ( Comment , :edit , where : [ user : [ role : :member ] ] ) Use allows to delegate permission to an association policy |&gt; allow ( Post , :read , where : [ archived : false ] ) |&gt; allow ( Comment , :read , where : [ post : allows ( :read ) ] ) Multiple allow combines as a logical-or # This will always allow reading all posts policy |&gt; allow ( Post , :read ) |&gt; allow ( Post , :read , where : [ archived : false ] ) # has no effect","ref":"basics.html#defining-rules","title":"The Basics - Defining rules","type":"extras"},{"doc":"Overview Hooks are attached to policies in the build_policy/2 callback. They are called prior to authorize or scope and can modify the resource/query or halt authorization altogether. Janus.Policy.attach_hook/4 Janus.Policy.attach_new_hook/4 Janus.Policy.detach_hook/3 Run prior to authorizing any schema policy |&gt; attach_hook ( :my_hook , fn :authorize , resource , _action -&gt; { :cont , resource } :scope , query , _action -&gt; { :cont , query } end ) Run prior to authorizing a specific schema policy |&gt; attach_hook ( :my_hook , Post , fn :authorize , resource , _action -&gt; { :cont , resource } :scope , query , _action -&gt; { :cont , query } end ) Used to preload fields policy |&gt; attach_hook ( :preload_user , Post , fn :authorize , post , _action -&gt; { :cont , Repo . preload ( post , :user ) } :scope , query , _action -&gt; { :cont , query } end ) Remove attached hooks with detach_hook/3 policy |&gt; attach_hook ( :my_hook , &amp; my_hook / 3 ) |&gt; detach_hook ( :my_hook ) policy |&gt; attach_hook ( :my_hook , Post , &amp; my_hook / 3 ) |&gt; detach_hook ( :my_hook , Post ) Attach hook if it's new using attach_new_hook/4 # second call has no effect policy |&gt; attach_new_hook ( :my_hook , &amp; my_hook / 3 ) |&gt; attach_new_hook ( :my_hook , &amp; my_other_hook / 3 ) # second call has no effect policy |&gt; attach_new_hook ( :my_hook , Post , &amp; my_hook / 3 ) |&gt; attach_new_hook ( :my_hook , Post , &amp; my_other_hook / 3 ) # attaches second hook because :my_hook not added for Post policy |&gt; attach_new_hook ( :my_hook , &amp; my_hook / 3 ) |&gt; attach_new_hook ( :my_hook , Post , &amp; my_other_hook / 3 ) Run a late check before each authorization call @impl true def build_policy ( policy , user ) do policy |&gt; attach_hook ( :ensure_unbanned , fn _ , object , _ -&gt; if Accounts . banned? ( user . id ) do :halt else { :cont , object } end end ) |&gt; ... end This can be useful if policies are being cached. If the call to Accounts.banned?(user.id) occurred in the callback body instead, the policy could not react to any change in account status after it was built.","ref":"basics.html#hooks","title":"The Basics - Hooks","type":"extras"},{"doc":"Pattern-match to give different permissions to different actors def build_policy ( policy , % User { role : :member } ) do # member permissions end def build_policy ( policy , % User { role : :moderator } ) do # moderator permissions end Delegate to context-specific policies def build_policy ( policy , actor ) do policy |&gt; CommunityForum.Policy . build_policy ( actor ) |&gt; Storefront.Policy . build_policy ( actor ) end For larger applications with well-defined boundaries, a policy can be constructed by threading it through multiple build_policy calls.","ref":"basics.html#structuring-your-policies","title":"The Basics - Structuring your policies","type":"extras"},{"doc":"","ref":"policy_usage.html","title":"Policy Usage","type":"extras"},{"doc":"Using policies in a context module Policies should most often be used in context modules, since they provide the interface to actions and resources that the rest of your application uses. defmodule MyApp.MessageBoard do @moduledoc &quot;&quot;&quot; Context module for the message board. &quot;&quot;&quot; import Ecto.Query # imports authorize, any_authorized?, scope, etc. import MyApp.Policy alias MyApp.Repo # ... end Authorizing an action using authorize/4 def update_post ( % Post { } = post , attrs \\\\ %{ } , user_or_policy ) do case authorize ( post , :edit , user_or_policy ) do { :ok , post } -&gt; post |&gt; Post . changeset ( attrs ) |&gt; Repo . update ( ) { :error , :not_authorized } -&gt; { :error , :not_authorized } end end iex&gt; update_post ( authorized_post , %{ } , some_user ) { :ok , % Post { } } iex&gt; update_post ( unauthorized_post , %{ } , some_user ) { :error , :not_authorized } Fetching authorized resources def authorized_posts ( user_or_policy ) do if any_authorized? ( Post , :read , user_or_policy ) do posts = Post |&gt; scope ( :read , user_or_policy ) |&gt; Repo . all ( ) { :ok , posts } else { :error , :not_authorized } end end Use any_authorized?/3 to differentiate between a result that is empty because there are no resources that match the policy conditions and a result that is empty because the user isn't authorized to view any resources. Preloading authorized associations scope ( Post , :read , user_or_policy , preload_authorized : :comments ) scope ( Post , :read , user_or_policy , preload_authorized : [ comments : :user ] ) The :preload_authorized option can be passed to preload only those associated resources that are authorized for the given action. Applying a query to preloads latest_comment_query = from Comment , order_by : [ desc : :inserted_at ] , limit : 1 scope ( Post , :read , user_or_policy , preload_authorized : [ comments : latest_comment_query ] ) A query can be applied to associated authorized resources. It is scoped per-association, so it applies to comments of each post instead of the comments of all posts. The above would return all :read -able posts preloaded with their latest :read -able comment. You can still include nested preloads using a tuple: scope ( Post , :read , user_or_policy , preload_authorized : [ comments : { latest_comment_query , [ :user ] } ] )","ref":"policy_usage.html#authorization","title":"Policy Usage - Authorization","type":"extras"},{"doc":"Call build_policy/1 to get the policy for a user iex&gt; policy = MyApp.Policy . build_policy ( current_user ) % Janus.Policy { ... } Pass a policy anywhere you'd pass in an actor iex&gt; MyApp.Policy . authorize ( post , :read , policy ) { :ok , post } iex&gt; MyApp.Policy . scope ( Post , :read , policy ) % Ecto.Query { } Cache a policy in a Plug.Conn Call in a plug def assign_current_policy ( conn ) do %{ assigns : %{ current_user : user } } = conn conn |&gt; assign ( :current_policy , MyApp.Policy . build_policy ( user ) ) end Controller action def index ( conn , _params ) do %{ assigns : %{ current_policy : policy } } = conn # Pass the policy to your context case MessageBoard . authorized_posts ( policy ) do { :ok , posts } -&gt; ... { :error , :not_authorized } -&gt; ... end end","ref":"policy_usage.html#caching-a-policy","title":"Policy Usage - Caching a policy","type":"extras"},{"doc":"Work in progress This guide is incomplete, though its contents may still be useful if you're getting started. Janus defines a small but flexible API that can be used to create a higher-level authorization interface for your application. To support common conventions, like the use of Phoenix-style context modules, Janus provides a policy generator that defines helpers that are especially useful in those circumstances. The goal of this guide is to explain the usage of, and reasoning behind, those helpers.","ref":"generated_policy_modules.html","title":"Generated Policy Modules","type":"extras"},{"doc":"The policy generator creates a policy module containing authorization functions that can be used by the rest of your application. $ mix janus.gen.policy [--module Example.Policy] [--path example/path/policy.ex] When run without arguments, it will generate a policy module called YourApp.Policy at lib/your_app/policy.ex (with YourApp replaced by your actual application namespace).","ref":"generated_policy_modules.html#generating-a-policy-module","title":"Generated Policy Modules - Generating a policy module","type":"extras"},{"doc":"Most of the generated helpers are designed to wrap common Ecto.Repo CRUD operations in a way that facilitates refactoring existing unauthorized calls to authorized ones with minimal necessary changes. The helpers don't cover everything, but they cover a large portion of use-cases and provide a pattern for wrapping additional operations if needed. The following helpers are included in newly-generated policy modules: authorized_fetch_by - wraps Ecto.Repo.get_by/3 , gets a resource using the given attributes and then authorizes for the given action/user authorized_fetch_all - wraps Ecto.Repo.all/2 , fetch a list of resources that are authorized for the given action/user authorized_insert - wraps Ecto.Repo.insert/2 , operates on a changeset, failing with a validation error if the inserted resource would not be authorized for the given action/user authorized_update - wraps Ecto.Repo.update/2 , operates on a changeset, failing with a validation error if the updated resource would not be authorized for the given action/user either before or after applying changes authorized_delete - wraps Ecto.Repo.delete/2 , deletes the given resource if it is authorized for the given action/user validate_authorized - changeset validation that ensures the resource being changed is authorized, adding a validation error otherwise. We'll go over each of these and how they might be used in the sections that follow. Note on generated code Remember that generated code is your code! It should be modified, replaced, and deleted as you see fit. Instead of including these helpers in Janus , they are generated as a starting point for your application. They enable the usage patterns described below, but if they aren't useful to you, toss them and use whatever you see fit!","ref":"generated_policy_modules.html#overview-of-helpers","title":"Generated Policy Modules - Overview of helpers","type":"extras"},{"doc":"The goal is to transition from unauthorized to authorized operations as smoothly as possible. authorized_* functions take the same arguments as their Ecto.Repo counterparts, except that they add additional keyword options related to authorization. Let's look at some examples. iex&gt; Repo . get_by ( Post , id : 12345 ) % Post { } iex&gt; Policy . authorized_fetch_by ( Post , [ id : 12345 ] , authorize : { :read , user } ) { :ok , % Post { } } # or { :error , :not_authorized } # or { :error , :not_found } All of the authorized_* functions take the :authorize keyword option, which should be either a tuple of {action, actor} as above, or false to skip authorization altogether. An ArgumentError will be raised if :authorize is not present, but this can be changed if you'd prefer authorization to be opt-in instead of opt-out. (Opt-out, the default, is the recommended approach, though it may require more refactoring work up-front.) There's another obvious difference in the example above: authorized_fetch_by/3 returns an :ok or :error tuple instead of the resource or nil . Returning a tuple allows us to differentiate between a lookup failure and an auth failure. A similar approach is used in the next example: iex&gt; Repo . all ( Post ) [ % Post { } , ... ] iex&gt; Policy . authorized_fetch_all ( Post , authorize : { :read , user } ) { :ok , [ % Post { } , ... ] } # or { :error , :not_authorized } A tuple is also returned here to differentiate authorization failures and an empty result. An Ecto query can also be passed as the first argument. iex&gt; query = from p in Post , where : p . inserted_at &gt; ago ( 1 , &quot;month&quot; ) iex&gt; Repo . all ( query ) [ % Post { } , ... ] iex&gt; Policy . authorized_fetch_all ( query , authorize : { :read , user } ) { :ok , [ % Post { } , ... ] } # or { :error , :not_authorized }","ref":"generated_policy_modules.html#authorized-operations","title":"Generated Policy Modules - Authorized operations","type":"extras"}]